#!/usr/bin/python2

import sys
import os
import taxtree

TreeFileName = sys.argv[1]
ReportFileName = None
fRep = None
if len(sys.argv) > 2:
	ReportFileName = sys.argv[2]
	fRep = open(ReportFileName, "a")

REFDIR = os.getenv("SERRATAX_REFDIR", "/serratax_ref")
if REFDIR.endswith("/"):
	REFDIR = REFDIR[:-1]

MetaFileName = REFDIR + "/doms.tsv"
CovTaxTreeFileName = REFDIR + "/cov.taxtree"

print >> sys.stderr, CovTaxTreeFileName
taxtree.Load(CovTaxTreeFileName)

'''
AlTega.memp.42  memp:HQ012367.25977.26861+ taxid=28295;sptaxid=28295;tax=superkingdom:Viruses,sub.superkingdom:Riboviria,kingdom:Orthornavirae,phylum:Pisuviricota,class:Pisoniviricetes,order:Nidovirales,suborder:Cornidovirineae,family:Coronaviridae,subfamily:Orthocoronavirinae,genus:Alphacoronavirus,subgenus:Pedacovirus,species:Porcine epidemic diarrhea virus;
'''

Ranks = [	
	"subfamily",
	"genus",
	"subgenus",
	"species",
	"sub.species",
#	"sub.no_rank",
	]

def TaxStrToDict(TaxStr):
	Fields = TaxStr.split(",")
	Dict = {}
	for Field in Fields:
		Rank, Name = Field.split(':')
		Dict[Rank] = Name
	return Dict
	
RefLabelToTaxStr = {}

for Line in open(MetaFileName):
	Fields = Line[:-1].split('\t')
	if len(Fields) != 2:
		sys.stderr.write(Fields)
		assert False
	RefLabel = Fields[0]
	Meta = Fields[1]
	n = Meta.find(' ')
	assert n > 0
	TaxStr = Meta[n+1:][:-1]
	RefLabelToTaxStr[RefLabel] = TaxStr

'''
#  0         1     2        3        4       5         6  7       8       9
Node    Parent  Left    Right   Length  Height  RootDist  Depth   Type    Label
0       *       1       30      *       0.484   0         0       Root    *
1       0       2       25      0.156   0.328   0.156     1       Int     *
2       1       3       4       0.238   0.0902  0.394     2       Int     *
3       2       *       *       0.0902  0       0.484     3       Leaf    AlTega.repl1a.18
4       2       5       6       0.0606  0.0296  0.454     3       Int     *
5       4       *       *       0.0296  0       0.484     4       Leaf    AlTega.repl1a.10
'''

Types = []
Lefts = []
Rights = []
Parents = []
Labels = []
Lengths = []
Heights = []
RootDists = []
Depths = []
TaxStrs = []
TaxDicts = []

f = open(TreeFileName)
Line = f.readline()
assert Line.startswith("Node\tParent")

QueryNode = None
NodeCount = 0
TreeLeafCount = 0
while True:
	Line = f.readline()
	if len(Line) == 0:
		break
	Fields = Line.split()
	assert len(Fields) == 10
	Node = int(Fields[0])
	assert Node == NodeCount

	Parent = None
	Left = None
	Right = None
	Label = None
	TaxStr = None
	TaxDict = None

	if Fields[1] != "*":
		Parent = int(Fields[1])
	if Fields[2] != "*":
		Left = int(Fields[2])
	if Fields[3] != "*":
		Right = int(Fields[3])
	Depth = int(Fields[7])

	Length = 0.0
	if NodeCount != 0:
		Length = float(Fields[4])
	Height = float(Fields[5])
	RootDist = float(Fields[6])
	Type = Fields[8]
	Label = Fields[9]
	if Type == "Leaf":
		TreeLeafCount += 1
		if Label.startswith("QORF"):
			assert QueryNode == None
			QueryNode = Node
		else:
			TaxStr = RefLabelToTaxStr[Label]
			TaxDict = TaxStrToDict(TaxStr)

	Types.append(Type)
	Lefts.append(Left)
	Rights.append(Right)
	Parents.append(Parent)
	Labels.append(Label)
	Lengths.append(Length)
	Heights.append(Height)
	RootDists.append(RootDist)
	Depths.append(Depth)
	TaxStrs.append(TaxStr)
	TaxDicts.append(TaxDict)
	NodeCount += 1
assert QueryNode != None

def GetSubtreeLeaves(Node):
	Leaves = [ ]
	PendingNodes = [ Node ]
	while len(PendingNodes) > 0:
		PendingNode = PendingNodes[0]
		PendingNodes.pop(0)
		if Types[PendingNode] == "Leaf":
			Leaves.append(PendingNode)
		else:
			Left = Lefts[PendingNode]
			Right = Rights[PendingNode]
			assert Left != None and Right != None
			PendingNodes.append(Left)
			PendingNodes.append(Right)
	return Leaves

RankNamePairToLeafCount = {}
for Node in range(0, NodeCount):
	if Types[Node] != "Leaf" or Node == QueryNode:
		continue
	D = TaxDicts[Node]
	for Rank in list(D.keys()):
		Name = D[Rank]
		Pair = (Rank, Name)
		try:
			RankNamePairToLeafCount[Pair] += 1
		except:
			RankNamePairToLeafCount[Pair] = 1

RankNamePairs = list(RankNamePairToLeafCount.keys())

NodeToRankNamePairToCount = []
for Node in range(0, NodeCount):
	Leaves = GetSubtreeLeaves(Node)
	RankNamePairToCount = {}
	RanksFound = []
	for Leaf in Leaves:
		if Leaf == QueryNode:
			continue
		D = TaxDicts[Leaf]
		for Rank in list(D.keys()):
			Name = D[Rank]
			Pair = (Rank, Name)
			try:
				RankNamePairToCount[Pair] += 1
			except:
				RankNamePairToCount[Pair] = 1

	NodeToRankNamePairToCount.append(RankNamePairToCount)

def GetLCA(RankNamePair):
	LCA = None
	MaxDepth = -1
	PairCount = RankNamePairToLeafCount[RankNamePair]
	for Node in range(0, NodeCount):
		if Node == QueryNode:
			continue
		PairToCount = NodeToRankNamePairToCount[Node]
		try:
			Count = PairToCount[RankNamePair]
		except:
			continue

		if Count == PairCount:
			Depth = Depths[Node]
			if Depth > MaxDepth:
				LCA = Node
				MaxDepth = Depth

	if LCA == None:
		print("No LCA for ", RankNamePair)
		assert False
	return LCA

def GetNames(Node, Rank):
	Names = []
	Leaves = GetSubtreeLeaves(Node)
	for Leaf in Leaves:
		if Leaf == QueryNode:
			continue
		D = TaxDicts[Leaf]
		RanksList = list(D.keys())
		for Rank2 in RanksList:
			if Rank2 == Rank:
				Name = D[Rank]
				if Name not in Names:
					Names.append(Name)
	return Names

PolyphyleticRankNamePairs = []
PolyCount = 0
for Rank in Ranks:
	for Pair in RankNamePairs:
		Rank2, Name = Pair
		if Rank2 != Rank:
			continue
		LCA = GetLCA(Pair)
		Names = GetNames(LCA, Rank)
		LeafCount = RankNamePairToLeafCount[Pair]
		s = "Rank=" + Rank
		s += ", name=" + Name
		s += ", %d leaves" % LeafCount
		s += ", LCA=" + str(LCA)
		if len(Names) == 1:
			s += ", monophyletic"
		else:
			PolyCount += 1
			for Name2 in Names:
				if Name2 != Name:
					s += ", polyphyletic(%s)" % Name2
					if Pair not in PolyphyleticRankNamePairs:
						PolyphyleticRankNamePairs.append(Pair)
					break
		print(s)

print("Polyphyletic=%d" % PolyCount)

QueryParent = Parents[QueryNode]
QueryParentDepth = Depths[QueryParent]
QueryParentHeight = Heights[QueryParent]
RankNamePairToCount = NodeToRankNamePairToCount[QueryParent]
s = "QueryParent=%d" % QueryParent
s += ", depth=%d" % QueryParentDepth
s += ", height=%.3g" % QueryParentHeight
print(s)

Pairs = list(RankNamePairToCount.keys())
QueryTaxStr = ""
LowestRank = ""
LowestName = ""
for Rank in Ranks:
	Names = []
	for Pair in Pairs:
		Rank2, Name = Pair
		if Rank2 != Rank:
			continue
		Names.append(Name)
	n = len(Names)
	QueryName = None
	s = "ParentRank=" + Rank
	s += ", names=%d" % n
	if n > 1:
		t = ""
		for Name in Names:
			if t != "":
				t += ";"
			t += Name
		s += "(" + t + ")"
	elif n == 1:
		Name = Names[0]
		Pair = (Rank, Name)
		if Pair in PolyphyleticRankNamePairs:
			s += ", %s:%s polyphyletic" % (Rank, Name)
			print(s)
			continue
		LCA = GetLCA(Pair)
		LCADepth = Depths[LCA]
		LCAHeight = Heights[LCA]
		s += ", lca_depth=%d" % LCADepth
		s += ", lca_height=%.3g" % LCAHeight
		if QueryParentDepth < LCADepth:
			s += "(below)"
		elif QueryParentHeight > LCAHeight*0.8:
			s += "(>0.8)"
		else:
			s += "(ok)"
			QueryName = Name
	if QueryName != None:
		s += ", queryname=" + QueryName
	print(s)
	if QueryName == None:
		break
	if QueryTaxStr != "":
		QueryTaxStr += ", "
	QueryTaxStr += Rank + ":" + Name
	LowestRank = Rank
	LowestName = Name

QueryTaxId = None
QueryTaxName = None
QueryRank = None
if LowestRank == "species" or LowestRank == "sub.species":
#	print("rank=[sub]species...")##@@
	for Node in taxtree.Nodes:
		if taxtree.NodeToName[Node] == LowestName:
			assert QueryTaxName == None
			QueryTaxName = LowestName
			QueryTaxId = Node
			QueryRank = taxtree.NodeToRank[Node]
			break
else:
	QueryTaxName = "unclassified " + LowestName
	Nodes = list(taxtree.NodeToName.keys())
	for Node in Nodes:
		Name2 = taxtree.NodeToName[Node]
		if Name2 == QueryTaxName:
			QueryTaxId = Node
			QueryRank = taxtree.NodeToRank[Node]
			break

print("QueryTaxStr=" + QueryTaxStr)
print("QueryTaxName=" + QueryTaxName)
print("QueryRank=" + str(QueryRank))
print("QueryTaxId=" + str(QueryTaxId))

if fRep != None:
	BaseName = TreeFileName.split('/')[-1]
	s = BaseName
	s += ", %u leaves" % TreeLeafCount
	s += ", %u polyphyletic clades" % PolyCount
	s += ", predict " + QueryTaxStr
	s += "\n"
	fRep.write(s)
