#!/usr/bin/python2

import sys
import os
import fasta
import random

'''
EU420137_10:25348_26022 nsp3:NC_032107.24614.25312+ subfamily:Orthocoronavirinae,genus:Alphacoronavirus,subgenus:Setracovirus,species:NL63-related bat coronavirus strain BtKYNL63-9b,sub.species:NL63-related bat coronavirus  1       217     225     96      1       219     233     94      39.8    3.02e-43
'''

OUTDIR = sys.argv[1]
ReportFileName = None
fRep = None
fDoms = open(OUTDIR + "/doms.txt", "w")
if len(sys.argv) > 2:
	ReportFileName = sys.argv[2]
	fRep = open(ReportFileName, "a")

REFDIR = os.getenv("SERRATAX_REFDIR", "/serratax_ref")
if REFDIR.endswith("/"):
	REFDIR = REFDIR[:-1]

RefFileName = REFDIR + "/doms.fa"
Domains = [ "memp", "nucp", "spike", "repl1a", "pol" ]

# TsvFileName is ublast output
#                   0      1   2   3  4
#	-userfields query+target+qlo+qhi+ql+qcov+tlo+thi+tl+tcov+id+evalue
# orfs2.fa is query ORF a.a. sequences from getorf

OrfFileName = OUTDIR + "/orfs2.fa"
TsvFileName = OUTDIR + "/ublast_orfs_domains.tsv"

RefSeqDict = fasta.ReadSeqsDict(RefFileName)
OrfSeqDict = fasta.ReadSeqsDict(OrfFileName)

# nAlDuvi.nsp4a.4
def GetDom(Label):
	Dom = Label.split('.')[1]
	return Dom

DomToORFs = {}
ORFToLength = {}
ORFToLines = {}
for Line in open(TsvFileName):
	Fields = Line[:-1].split('\t')
	ORF = Fields[0]
	RefLabel = Fields[1]
	Dom = GetDom(RefLabel)
	if not Dom in Domains:
		continue
	ORFLength = int(Fields[4])
	ORFToLength[ORF] = ORFLength

	try:
		v = ORFToLines[ORF]
	except:
		ORFToLines[ORF] = []
	ORFToLines[ORF].append(Line)

	try:
		v = DomToORFs[Dom]
	except:
		DomToORFs[Dom] = [ ]
	if ORF not in DomToORFs[Dom]:
		DomToORFs[Dom].append(ORF)

if 0:
	print >> sys.stderr, "DomToORFs:"
	print >> sys.stderr, DomToORFs

def OutputDom(Dom, ORF):
	Lines = ORFToLines[ORF]
	if len(Lines) < 3:
		return

	fFa = open(OUTDIR + "/" + Dom + ".fa", "w")

	ORFSeq = OrfSeqDict[ORF]
	fasta.WriteSeq(fFa, ORFSeq, ORF)

	fORF = open(OUTDIR + "/" + "orf." + Dom + ".fa", "w")
	fasta.WriteSeq(fORF, ORFSeq, ORF)
	fORF.close()

	SelectedLines = []
	OtherLines = []
	N = len(Lines)
	for i in range(0, N):
		if i < 16:
			SelectedLines.append(Lines[i])
		else:
			OtherLines.append(Lines[i])

	random.shuffle(OtherLines)
	M = len(OtherLines)
	if M > 16:
		M = 16
	for i in range(0, M):
		SelectedLines.append(OtherLines[i])

	FoundRefLabels = []
	Total = 0
	for Line in SelectedLines:
		Fields = Line[:-1].split('\t')
		ORFLabel = Fields[0]
		RefLabel = Fields[1]
		Dom2 = GetDom(RefLabel)
		if Dom2 != Dom:
			continue
		## if RefLabel in FoundRefLabels:
		##	continue
		FoundRefLabels.append(RefLabel)
		RefSeq = RefSeqDict[RefLabel]
		fasta.WriteSeq(fFa, RefSeq, RefLabel)
		Total += 1
	fFa.close()
	if Total < 8:
		return

	fDoms.write(Dom + "\n")
	if fRep != None:
		s = "%s assigned to %s" % (ORF, Dom)
		fRep.write(s + "\n")

for Dom in list(DomToORFs.keys()):
	try:
		ORFs = DomToORFs[Dom]
	except:
		continue

	MaxL = 0
	BestORF = None
	for ORF in ORFs:
		L = ORFToLength[ORF]
		if L > MaxL:
			MaxL = L
			BestORF = ORF
	assert BestORF != None
	OutputDom(Dom, BestORF)
